package io.github.null2264.tsukumogami.core

import co.touchlab.kermit.Logger
import dev.kord.core.Kord
import dev.kord.core.entity.Message
import dev.kord.core.event.gateway.ReadyEvent
import dev.kord.core.event.message.MessageCreateEvent
import dev.kord.core.on
import dev.kord.gateway.Intent
import dev.kord.gateway.PrivilegedIntent
import io.github.null2264.tsukumogami.core.commands.Command
import io.github.null2264.tsukumogami.core.exceptions.CommandException
import io.github.null2264.tsukumogami.core.exceptions.CommandNotFound
import io.github.null2264.tsukumogami.core.module.BotModule
import io.github.null2264.tsukumogami.core.commands.IGroup
import io.github.null2264.tsukumogami.core.ext.parseCommandAndArguments
import kotlin.reflect.full.callSuspend
import kotlinx.coroutines.runBlocking

open class Bot internal constructor(): IGroup {

    lateinit var client: Kord
    private val modules = mutableMapOf<String, BotModule>()
    private val _prefixes = mutableListOf<String>()
    val prefixes: List<String> get() = _prefixes.toList()
    internal lateinit var token: String
    override val allCommands: MutableMap<String, Command> = mutableMapOf()

    fun addModule(module: BotModule) {
        modules[module.name] = module.install(this)
    }

    fun addPrefix(prefix: String) {
        _prefixes.add(prefix)
    }

    fun addCommand(command: Command) {
        if (allCommands.containsKey(command.name)) {
            throw IllegalStateException("Duplicate command: '${command.name}'")
        }
        allCommands[command.name] = command
    }

    fun removeCommand(command: Command) {
        if (!allCommands.containsKey(command.name)) {
            throw IllegalStateException("Command not found: '${command.name}'")
        }
        allCommands.remove(command.name)
    }

    open suspend fun start() {
        client = Kord(token).apply {
            on<ReadyEvent> { onReady() }
            on<MessageCreateEvent> { onMessage(this.message, this) }
        }

        client.login {
            @OptIn(PrivilegedIntent::class)
            intents += Intent.MessageContent
        }
    }

    open suspend fun stop() {
        client.shutdown()
    }

    fun getCommand(name: String?) = allCommands[name]

    private fun getContext(message: Message): Context {
        val candidate = message.content.parsePrefixCommandAndArguments()
        val context = Context(this, message, candidate?.first, candidate?.second)
        context.command = getCommand(candidate?.second?.get(0))
        return context
    }

    open suspend fun onCommandError(context: Context, error: CommandException) {
        context.send(error.message!!)
    }

    open suspend fun processCommand(message: Message) {
        val ctx = getContext(message)

        try {
            ctx.command?.invoke(ctx) ?: throw CommandNotFound()
        } catch (e: CommandException) {
            onCommandError(ctx, e)
        } catch (e: Exception) {
            Logger.e(e) { "Something went wrong while trying to process command '${ctx.command?.name}'" }
        }
    }

    open suspend fun onMessage(message: Message, event: MessageCreateEvent) {
        if (message.author?.isBot != false) return

        processCommand(message)
    }

    open suspend fun onReady() {
        Logger.i { "Online! ${client.getSelf().username}" }
    }

    private fun String.parsePrefixCommandAndArguments(): Pair<String, List<String>>? {
        if (this.isBlank()) return null

        var ret: Pair<String, List<String>>? = null

        prefixes.forEach {
            if (this.substring(0, it.length) == it) {
                val prefix = this.substring(0, it.length)
                val cleanPrompt = this.substring(it.length)
                ret = Pair(prefix, cleanPrompt.parseCommandAndArguments())
                return@forEach
            }
        }

        return ret
    }
}
